import importlib
import os
from contextlib import suppress
from functools import cached_property
from typing import Any, get_type_hints

from pyttings.type_converter import convert_and_validate


class Settings:
    def __init__(self) -> None:
        """Initialize the settings manager."""
        self._settings_module: str = self._load_settings_module()
        self._env_prefix: str = os.getenv("PYTTING_ENV_PREFIX", "PYTTING_")
        self._cache: dict[str, Any] = {}

    def _load_settings_module(self) -> str:
        """Get the settings module name from environment variable."""
        settings_module: str | None = os.getenv("PYTTING_SETTINGS_MODULE")
        if settings_module is None:
            raise ValueError(
                "'PYTTING_SETTINGS_MODULE' environment variable is not set.\n"
                "Please specify a settings module."
            )
        return settings_module

    @cached_property
    def _module(self):
        """Import and cache the settings module."""
        return importlib.import_module(self._settings_module)

    @cached_property
    def _type_hints(self) -> dict[str, Any]:
        """Get type hints from the settings module."""
        with suppress(ImportError, ValueError, TypeError):
            return get_type_hints(self._module)
        return {}

    @cached_property
    def defaults(self) -> dict[str, Any]:
        """Get all uppercase attributes from the settings module as defaults."""
        return {
            key: getattr(self._module, key)
            for key in dir(self._module)
            if not key.startswith("__") and not key.endswith("__") and key.isupper()
        }

    def get_env_var(self, name: str) -> Any | None:
        """Get and convert environment variable for a setting."""
        env_var_name = f"{self._env_prefix}{name}"
        value = os.getenv(env_var_name)

        if value is None or name not in self.defaults:
            return value
        expected_type = self._type_hints.get(name, type(self.defaults[name]))
        return convert_and_validate(name, value, expected_type)

    def __getattr__(self, name: str) -> Any:
        if name not in self._cache:
            value: Any | None = self.get_env_var(name)
            if value is None:
                if name in self.defaults:
                    value = self.defaults[name]
                else:
                    raise AttributeError(
                        f"'{self.__class__.__name__}' object has no attribute '{name}'"
                    )
            self._cache[name] = value
        return self._cache[name]



import ast
import inspect
import os
import types
from contextlib import suppress
from typing import Any, Type, TypeVar, Union, get_args, get_origin

from pyttings.exceptions import SettingMisconfigured

CollectionT = TypeVar("CollectionT", list, tuple, set, dict)
CONTAINER_TYPES = {list, tuple, set, dict}
UNION_TYPES = {types.UnionType, Union}

CUSTOM_CLASS_METHOD_NAME = os.getenv(
    "PYTTING_CUSTOM_CLASS_METHOD_NAME", "__pyttings_convert__"
)


def is_custom_class(expected_type: type) -> bool:
    """Check if a type has a custom conversion method."""
    return hasattr(expected_type, CUSTOM_CLASS_METHOD_NAME) and callable(
        getattr(expected_type, CUSTOM_CLASS_METHOD_NAME)
    )


def convert_container(
    value: str, expected_type: Type[CollectionT]
) -> CollectionT | None:
    """Try to convert a string to a container type using ast.literal_eval."""
    with suppress(SyntaxError, ValueError, TypeError):
        parsed_value = ast.literal_eval(value)
        return (
            expected_type(parsed_value)
            if isinstance(parsed_value, expected_type)
            else None
        )
    return None


def validate_container_types(
    value: Any, container_type: Type[CollectionT], arg_types: tuple | None
) -> bool:
    """Validate that container elements match their expected types."""
    if arg_types is None:
        return True

    if container_type is dict:
        if not isinstance(value, dict):
            return False
        key_type, value_type = arg_types
        return all(
            isinstance(k, key_type) and isinstance(v, value_type)
            for k, v in value.items()
        )

    if container_type in {list, tuple, set}:
        if not isinstance(value, container_type):
            return False
        element_type = arg_types[0]
        return all(isinstance(x, element_type) for x in value)

    return False


def handle_union_type(name: str, value: str, union_type: type) -> Any:
    """Try to convert the value to any of the types in the union."""
    for candidate_type in get_args(union_type):
        with suppress(SettingMisconfigured):
            return convert_and_validate(name, value, candidate_type)

    raise SettingMisconfigured(
        f"Invalid type for {name} with configured value '{value}'. "
        f"Expected one of {get_args(union_type)}."
    )


def handle_custom_class(name: str, value: str, cls_type: type) -> Any:
    """Handle conversion for custom classes with conversion methods."""
    method = getattr(cls_type, CUSTOM_CLASS_METHOD_NAME)
    params = list(inspect.signature(method).parameters.values())

    if len(params) != 1:
        raise SettingMisconfigured(
            f"Invalid method signature for {cls_type}. Expected a single parameter."
        )

    param = params[0]
    if param.annotation is inspect.Parameter.empty:
        raise SettingMisconfigured(
            f"Invalid method signature for {cls_type}. "
            f"Expected a type hint for the parameter."
        )

    converted_param = convert_and_validate(name, value, param.annotation)
    return method(converted_param)


def handle_simple_type(name: str, value: str, expected_type: type) -> Any:
    """Handle conversion for simple, non-generic types."""
    if expected_type is bool:
        return value.lower() == "true"
    elif expected_type is types.NoneType:
        return value
    elif expected_type in CONTAINER_TYPES:
        converted_value = convert_container(value, expected_type)
        if converted_value is not None:
            return converted_value
    else:
        with suppress(ValueError, TypeError):
            return expected_type(value)

    raise SettingMisconfigured(
        f"Invalid type for {name} with configured value '{value}'."
        f"\nExpected {expected_type}."
    )


def handle_generic_container(
    name: str, value: str, origin: type, expected_type: type
) -> Any:
    """Handle conversion for generic container types."""
    converted_value = convert_container(value, origin)
    if converted_value is not None and validate_container_types(
        converted_value, origin, get_args(expected_type)
    ):
        return converted_value

    raise SettingMisconfigured(
        f"Invalid type for {name} with configured value '{value}'."
        f"\nExpected {expected_type}."
    )


def convert_and_validate(name: str, value: str, expected_type: type) -> Any:
    origin = get_origin(expected_type)

    if origin in UNION_TYPES:
        return handle_union_type(name, value, expected_type)

    if origin is None:
        if is_custom_class(expected_type):
            return handle_custom_class(name, value, expected_type)

        return handle_simple_type(name, value, expected_type)
    elif origin in CONTAINER_TYPES:
        return handle_generic_container(name, value, origin, expected_type)

    raise SettingMisconfigured(
        f"Invalid type for {name} with configured value '{value}'."
        f"\nExpected {expected_type}."
    )



import os
from decimal import Decimal, InvalidOperation

import pytest

from pyttings import settings
from pyttings.core import Settings
from pyttings.exceptions import SettingMisconfigured
from tests.utils import ListOfInts, MultipleArgsCustomClass


@pytest.fixture(autouse=True)
def clear_settings_cache():
    settings._cache = {}


# Test missing settings module
def test_missing_settings_module():
    del os.environ["PYTTING_SETTINGS_MODULE"]
    with pytest.raises(
        ValueError,
        match="'PYTTING_SETTINGS_MODULE' environment variable is not set.\nPlease specify a settings module.",
    ):
        Settings()


# Test missing attribute
def test_missing_attribute():
    with pytest.raises(AttributeError, match="has no attribute 'MISSING_SETTING'"):
        _ = settings.MISSING_SETTING


# Test custom prefix
def test_custom_prefix():
    os.environ["PYTTING_ENV_PREFIX"] = "TEST_"
    os.environ["PYTTING_SETTINGS_MODULE"] = "tests.settings"
    os.environ["TEST_OTHER_SETTING"] = "test_value"
    new_settings = Settings()
    assert new_settings._env_prefix == "TEST_"
    assert new_settings.DEBUG is True
    assert new_settings.DATABASE_URL == "sqlite:///db.sqlite3"
    assert new_settings.OTHER_SETTING == "test_value"


# Test basic settings initialization
def test_settings_basic_values():
    assert settings.DEBUG is True
    assert settings.DATABASE_URL == "sqlite:///db.sqlite3"
    assert settings.SECRET_KEY == "my-secret-key"
    assert settings.PORT == 8000
    assert settings.ENABLE_FEATURE is False


def test_settings_collections():
    assert settings.SOME_LIST == ["a", "b", "c"]
    assert settings.SOME_TUPLE == ("a", "b", "c")
    assert settings.SOME_SET == {"a", "b", "c"}
    assert settings.SOME_DICT == {"a": "b", "c": "d"}


def test_settings_special_types():
    assert settings.NONE_VALUE is None
    assert settings.SOME_UNION_TYPE == "some_str"
    assert settings.SOME_DECIMAL == Decimal("1.0")
    assert settings.SOME_CUSTOM_CLASS == ListOfInts([1, 2, 3])
    assert settings.SOME_MULTIPLE_CUSTOM_CLASS == MultipleArgsCustomClass(1, "2", 3)


def test_settings_no_type_hints():
    assert settings.NO_TYPE_HINT_NONE is None
    assert settings.NO_TYPE_HINT_BOOL is True
    assert settings.NO_TYPE_HINT_INT == 1
    assert settings.NO_TYPE_HINT_FLOAT == 1.0
    assert settings.NO_TYPE_HINT_STR == "a"


def test_settings_no_type_hints_collections():
    assert settings.NO_TYPE_HINT_DICT == {"a": "b", "c": "d"}
    assert settings.NO_TYPE_HINT_LIST == ["a", "b", "c"]
    assert settings.NO_TYPE_HINT_TUPLE == ("a", "b", "c")
    assert settings.NO_TYPE_HINT_SET == {"a", "b", "c"}
    assert settings.NO_TYPE_HINT_DECIMAL == Decimal("1.0")


# Test environment variable overrides - basic types
def test_env_var_overrides_basic():
    os.environ["PYTTING_DEBUG"] = "False"
    os.environ["PYTTING_DATABASE_URL"] = "postgresql://user:pass@localhost/db"
    os.environ["PYTTING_SECRET_KEY"] = "new-secret-key"
    os.environ["PYTTING_PORT"] = "8080"
    os.environ["PYTTING_ENABLE_FEATURE"] = "True"

    assert settings.DEBUG is False
    assert settings.DATABASE_URL == "postgresql://user:pass@localhost/db"
    assert settings.SECRET_KEY == "new-secret-key"
    assert settings.PORT == 8080
    assert settings.ENABLE_FEATURE is True


# Test environment variable overrides - collections
def test_env_var_overrides_collections():
    os.environ["PYTTING_SOME_LIST"] = '["x", "y"]'
    os.environ["PYTTING_SOME_TUPLE"] = '("x", "y")'
    os.environ["PYTTING_SOME_SET"] = '{"x", "y"}'
    os.environ["PYTTING_SOME_DICT"] = '{"x": "y"}'

    assert settings.SOME_LIST == ["x", "y"]
    assert settings.SOME_TUPLE == ("x", "y")
    assert settings.SOME_SET == {"x", "y"}
    assert settings.SOME_DICT == {"x": "y"}


# Test environment variable overrides - special types
def test_env_var_overrides_special_types():
    os.environ["PYTTING_NONE_VALUE"] = "some_value"
    os.environ["PYTTING_SOME_UNION_TYPE"] = "some_other_str"
    os.environ["PYTTING_SOME_DECIMAL"] = "2.0"

    assert settings.NONE_VALUE == "some_value"
    assert settings.SOME_UNION_TYPE == "some_other_str"
    assert settings.SOME_DECIMAL == Decimal("2.0")


# Test environment variable overrides - custom classes
def test_env_var_overrides_custom_classes():
    os.environ["PYTTING_SOME_CUSTOM_CLASS"] = "[1, 2, 3, 4]"
    os.environ["PYTTING_SOME_MULTIPLE_CUSTOM_CLASS"] = (
        '{"int_value": 3, "str_value": "2", "value": "1"}'
    )

    assert settings.SOME_CUSTOM_CLASS == ListOfInts([1, 2, 3, 4])
    assert settings.SOME_MULTIPLE_CUSTOM_CLASS == MultipleArgsCustomClass(3, "2", "1")


# Test environment variable overrides - no type hints
def test_env_var_overrides_no_type_hint_simple():
    os.environ["PYTTING_NO_TYPE_HINT_NONE"] = "tests.settings"
    os.environ["PYTTING_NO_TYPE_HINT_BOOL"] = "False"
    os.environ["PYTTING_NO_TYPE_HINT_INT"] = "42"
    os.environ["PYTTING_NO_TYPE_HINT_FLOAT"] = "3.14"
    os.environ["PYTTING_NO_TYPE_HINT_STR"] = "override"

    assert settings.NO_TYPE_HINT_NONE == "tests.settings"
    assert settings.NO_TYPE_HINT_BOOL is False
    assert settings.NO_TYPE_HINT_INT == 42
    assert settings.NO_TYPE_HINT_FLOAT == 3.14
    assert settings.NO_TYPE_HINT_STR == "override"


def test_env_var_overrides_no_type_hint_collections():
    os.environ["PYTTING_NO_TYPE_HINT_DICT"] = '{"x": "y"}'
    os.environ["PYTTING_NO_TYPE_HINT_LIST"] = '["x", "y"]'
    os.environ["PYTTING_NO_TYPE_HINT_TUPLE"] = '("x", "y")'
    os.environ["PYTTING_NO_TYPE_HINT_SET"] = '{"x", "y"}'
    os.environ["PYTTING_NO_TYPE_HINT_DECIMAL"] = "3.14"

    assert settings.NO_TYPE_HINT_DICT == {"x": "y"}
    assert settings.NO_TYPE_HINT_LIST == ["x", "y"]
    assert settings.NO_TYPE_HINT_TUPLE == ("x", "y")
    assert settings.NO_TYPE_HINT_SET == {"x", "y"}
    assert settings.NO_TYPE_HINT_DECIMAL == Decimal("3.14")


# Test complex environment variable type conversions
def test_env_var_type_conversion_collections():
    os.environ["PYTTING_SOME_LIST"] = '["a", "b", "c", "d"]'
    os.environ["PYTTING_SOME_TUPLE"] = "(1,2,3)"
    os.environ["PYTTING_SOME_SET"] = '{"a", "b", "c", 1, 2, 3}'

    assert settings.SOME_LIST == ["a", "b", "c", "d"]
    assert settings.SOME_TUPLE == (1, 2, 3)
    assert settings.SOME_SET == {"a", "b", "c", 1, 2, 3}


def test_env_var_type_conversion_complex_dict():
    os.environ["PYTTING_SOME_DICT"] = (
        '{"a": "b", "c": "d", "another_dict": {"a": 1, "b": 2, "c": 3}}'
    )

    assert settings.SOME_DICT == {
        "a": "b",
        "c": "d",
        "another_dict": {"a": 1, "b": 2, "c": 3},
    }


# Test union type conversion
def test_env_var_type_conversion_union_str():
    os.environ["PYTTING_SOME_UNION_TYPE"] = "some_other_str"
    assert settings.SOME_UNION_TYPE == "some_other_str"


def test_env_var_type_conversion_union_dict():
    settings._cache.pop("SOME_UNION_TYPE", None)
    os.environ["PYTTING_SOME_UNION_TYPE"] = '{"x": "y"}'
    assert settings.SOME_UNION_TYPE == {"x": "y"}


# Test strict type checking
def test_env_var_type_conversion_strict_valid():
    os.environ["PYTTING_SOME_STRICT_DICT"] = '{"x": "y"}'
    os.environ["PYTTING_SOME_STRICT_LIST"] = '["x", "y"]'

    assert settings.SOME_STRICT_DICT == {"x": "y"}
    assert settings.SOME_STRICT_LIST == ["x", "y"]


def test_env_var_type_conversion_strict_list_invalid():
    settings._cache.pop("SOME_STRICT_LIST", None)
    os.environ["PYTTING_SOME_STRICT_LIST"] = "[1, 'x', 'y']"

    with pytest.raises(
        SettingMisconfigured,
        match=r"Invalid type for SOME_STRICT_LIST with configured value '.*'\.\nExpected list\[str\]\.",
    ):
        _ = settings.SOME_STRICT_LIST


def test_env_var_type_conversion_strict_dict_invalid_key():
    settings._cache.pop("SOME_STRICT_DICT", None)
    os.environ["PYTTING_SOME_STRICT_DICT"] = "{1: 1}"

    with pytest.raises(
        SettingMisconfigured,
        match=r"Invalid type for SOME_STRICT_DICT with configured value '.*'\.\nExpected dict\[str, str\].",
    ):
        _ = settings.SOME_STRICT_DICT


def test_env_var_type_conversion_strict_dict_invalid_value():
    settings._cache.pop("SOME_STRICT_DICT", None)
    os.environ["PYTTING_SOME_STRICT_DICT"] = "{'x': 1}"

    with pytest.raises(
        SettingMisconfigured,
        match=r"Invalid type for SOME_STRICT_DICT with configured value '.*'\.\nExpected dict\[str, str\].",
    ):
        _ = settings.SOME_STRICT_DICT


def test_env_var_type_conversion_strict_dict_invalid_key_type():
    settings._cache.pop("SOME_STRICT_DICT", None)
    os.environ["PYTTING_SOME_STRICT_DICT"] = "{1: 'x'}"

    with pytest.raises(
        SettingMisconfigured,
        match=r"Invalid type for SOME_STRICT_DICT with configured value '.*'\.\nExpected dict\[str, str\].",
    ):
        _ = settings.SOME_STRICT_DICT


# Test custom class validation
def test_env_var_type_conversion_strict_custom_class_invalid_list():
    os.environ["PYTTING_SOME_CUSTOM_CLASS"] = '[1, 2, 3, "4"]'

    with pytest.raises(
        SettingMisconfigured,
        match=r"Invalid type for SOME_CUSTOM_CLASS with configured value '.*'\.\nExpected list\[int\]\.",
    ):
        _ = settings.SOME_CUSTOM_CLASS


def test_env_var_type_conversion_strict_custom_class_invalid_dict():
    os.environ["PYTTING_SOME_MULTIPLE_CUSTOM_CLASS"] = (
        '{"int_value": [1], "str_value": "2", "value": "1"}'
    )

    with pytest.raises(
        SettingMisconfigured,
        match=r"Invalid type for SOME_MULTIPLE_CUSTOM_CLASS with configured value '.*'\.\nExpected dict\[str, int \| str\]\.",
    ):
        _ = settings.SOME_MULTIPLE_CUSTOM_CLASS


# Test type conversion failures
def test_env_var_type_conversion_failure_list():
    os.environ["PYTTING_SOME_LIST"] = "1"

    with pytest.raises(
        SettingMisconfigured,
        match="Invalid type for SOME_LIST with configured value '1'.\nExpected <class 'list'>.",
    ):
        _ = settings.SOME_LIST


def test_env_var_type_conversion_failure_tuple():
    os.environ["PYTTING_SOME_TUPLE"] = "a"

    with pytest.raises(
        SettingMisconfigured,
        match="Invalid type for SOME_TUPLE with configured value 'a'.\nExpected <class 'tuple'>.",
    ):
        _ = settings.SOME_TUPLE


def test_env_var_type_conversion_failure_int():
    os.environ["PYTTING_PORT"] = ""

    with pytest.raises(
        SettingMisconfigured,
        match="Invalid type for PORT with configured value ''.\nExpected <class 'int'>.",
    ):
        _ = settings.PORT


def test_env_var_type_conversion_failure_dict():
    os.environ["PYTTING_NO_TYPE_HINT_DICT"] = "123"

    with pytest.raises(
        SettingMisconfigured,
        match="Invalid type for NO_TYPE_HINT_DICT with configured value '123'.\nExpected <class 'dict'>.",
    ):
        _ = settings.NO_TYPE_HINT_DICT


def test_env_var_type_conversion_failure_decimal():
    os.environ["PYTTING_NO_TYPE_HINT_DECIMAL"] = "A"

    with pytest.raises(InvalidOperation):
        _ = settings.NO_TYPE_HINT_DECIMAL



import os
import types
from decimal import Decimal
from importlib import reload
from typing import Dict, List, Set, Tuple, Union

import pytest

from pyttings.exceptions import SettingMisconfigured
from pyttings.type_converter import (
    convert_and_validate,
    convert_container,
    is_custom_class,
    validate_container_types,
)


class SimpleCustomClass:
    def __init__(self, value):
        self.value = value

    @classmethod
    def __pyttings_convert__(cls, value: List[int]):
        return cls(value)

    def __eq__(self, other):
        if not isinstance(other, SimpleCustomClass):
            return False
        return self.value == other.value


class InvalidCustomClass:
    @classmethod
    def __pyttings_convert__(cls):
        # Missing parameter
        return cls()


class UntypedCustomClass:
    @classmethod
    def __pyttings_convert__(cls, value):
        # Missing type hint
        return cls(value)

    def __init__(self, value):
        self.value = value


# Test is_custom_class function
def test_is_custom_class():
    assert is_custom_class(SimpleCustomClass) is True
    assert is_custom_class(int) is False
    assert is_custom_class(str) is False

    # Test with different method name
    original_method_name = os.environ.get("PYTTING_CUSTOM_CLASS_METHOD_NAME")
    try:
        os.environ["PYTTING_CUSTOM_CLASS_METHOD_NAME"] = "custom_convert"

        class CustomMethodClass:
            @classmethod
            def custom_convert(cls, value: int):
                return cls(value)

            def __init__(self, value):
                self.value = value

        import pyttings.type_converter

        reload(pyttings.type_converter)

        assert pyttings.type_converter.is_custom_class(CustomMethodClass) is True
        assert pyttings.type_converter.is_custom_class(SimpleCustomClass) is False
    finally:
        if original_method_name is None:
            del os.environ["PYTTING_CUSTOM_CLASS_METHOD_NAME"]
        else:
            os.environ["PYTTING_CUSTOM_CLASS_METHOD_NAME"] = original_method_name
        reload(pyttings.type_converter)


# Test convert_container function
def test_convert_container_list():
    # Valid list
    assert convert_container("[1, 2, 3]", list) == [1, 2, 3]
    # Valid list with strings
    assert convert_container('["a", "b", "c"]', list) == ["a", "b", "c"]
    # Invalid syntax
    assert convert_container("[1, 2, 3", list) is None
    # Wrong type
    assert convert_container('{"a": 1}', list) is None


def test_convert_container_tuple():
    # Valid tuple
    assert convert_container("(1, 2, 3)", tuple) == (1, 2, 3)
    # Valid tuple with strings
    assert convert_container('("a", "b", "c")', tuple) == ("a", "b", "c")
    # Invalid syntax
    assert convert_container("(1, 2, 3", tuple) is None
    # Wrong type
    assert convert_container("[1, 2, 3]", tuple) is None


def test_convert_container_set():
    # Valid set
    assert convert_container("{1, 2, 3}", set) == {1, 2, 3}
    # Valid set with strings
    assert convert_container('{"a", "b", "c"}', set) == {"a", "b", "c"}
    # Invalid syntax
    assert convert_container("{1, 2, 3", set) is None
    # Wrong type
    assert convert_container("[1, 2, 3]", set) is None


def test_convert_container_dict():
    # Valid dict
    assert convert_container('{"a": 1, "b": 2}', dict) == {"a": 1, "b": 2}
    # Valid dict with nested structures
    assert convert_container('{"a": [1, 2], "b": {"c": 3}}', dict) == {
        "a": [1, 2],
        "b": {"c": 3},
    }
    # Invalid syntax
    assert convert_container('{"a": 1, "b": 2', dict) is None
    # Wrong type
    assert convert_container("[1, 2, 3]", dict) is None


# Test validate_container_types function
def test_validate_container_types_list():
    # Valid homogeneous list
    assert validate_container_types([1, 2, 3], list, (int,)) is True
    # Valid mixed list (when type checking is not enforced)
    assert validate_container_types([1, "a", 3.0], list, None) is True
    # Invalid type
    assert validate_container_types([1, "a", 3], list, (int,)) is False
    # Not a list
    assert validate_container_types({1, 2, 3}, list, (int,)) is False


def test_validate_container_types_tuple():
    # Valid homogeneous tuple
    assert validate_container_types((1, 2, 3), tuple, (int,)) is True
    # Valid mixed tuple (when type checking is not enforced)
    assert validate_container_types((1, "a", 3.0), tuple, None) is True
    # Invalid type
    assert validate_container_types((1, "a", 3), tuple, (int,)) is False
    # Not a tuple
    assert validate_container_types([1, 2, 3], tuple, (int,)) is False


def test_validate_container_types_set():
    # Valid homogeneous set
    assert validate_container_types({1, 2, 3}, set, (int,)) is True
    # Valid mixed set (when type checking is not enforced)
    assert validate_container_types({1, "a", 3.0}, set, None) is True
    # Invalid type
    assert validate_container_types({1, "a", 3}, set, (int,)) is False
    # Not a set
    assert validate_container_types([1, 2, 3], set, (int,)) is False


def test_validate_container_types_dict():
    # Valid homogeneous dict
    assert validate_container_types({"a": 1, "b": 2}, dict, (str, int)) is True
    # Valid mixed dict (when type checking is not enforced)
    assert validate_container_types({"a": 1, 2: "b"}, dict, None) is True
    # Invalid key type
    assert validate_container_types({1: "a", 2: "b"}, dict, (str, str)) is False
    # Invalid value type
    assert validate_container_types({"a": 1, "b": 2}, dict, (str, str)) is False
    # Not a dict
    assert validate_container_types([1, 2, 3], dict, (str, int)) is False


# Test convert_and_validate function
def test_convert_and_validate_simple_types():
    # Integer
    assert convert_and_validate("TEST_INT", "42", int) == 42
    # String
    assert convert_and_validate("TEST_STR", "hello", str) == "hello"
    # Boolean - true
    assert convert_and_validate("TEST_BOOL", "true", bool) is True
    # Boolean - false
    assert convert_and_validate("TEST_BOOL", "false", bool) is False
    # Boolean - case insensitive
    assert convert_and_validate("TEST_BOOL", "TRUE", bool) is True
    # Float
    assert convert_and_validate("TEST_FLOAT", "3.14", float) == 3.14
    # Decimal
    assert convert_and_validate("TEST_DECIMAL", "3.14", Decimal) == Decimal("3.14")
    # None - returns the string value
    assert convert_and_validate("TEST_NONE", "null", types.NoneType) == "null"


def test_convert_and_validate_failures():
    # Invalid int
    with pytest.raises(
        SettingMisconfigured,
        match="Invalid type for TEST_INT with configured value 'abc'",
    ):
        convert_and_validate("TEST_INT", "abc", int)

    # Invalid float
    with pytest.raises(
        SettingMisconfigured,
        match="Invalid type for TEST_FLOAT with configured value 'xyz'",
    ):
        convert_and_validate("TEST_FLOAT", "xyz", float)


def test_convert_and_validate_containers():
    # List
    assert convert_and_validate("TEST_LIST", "[1, 2, 3]", list) == [1, 2, 3]
    # Tuple
    assert convert_and_validate("TEST_TUPLE", "(1, 2, 3)", tuple) == (1, 2, 3)
    # Set
    assert convert_and_validate("TEST_SET", "{1, 2, 3}", set) == {1, 2, 3}
    # Dict
    assert convert_and_validate("TEST_DICT", '{"a": 1, "b": 2}', dict) == {
        "a": 1,
        "b": 2,
    }


def test_convert_and_validate_typed_containers():
    # List[int]
    list_int_type = List[int]
    assert convert_and_validate("TEST_LIST", "[1, 2, 3]", list_int_type) == [1, 2, 3]

    # List[str]
    list_str_type = List[str]
    assert convert_and_validate("TEST_LIST", '["a", "b", "c"]', list_str_type) == [
        "a",
        "b",
        "c",
    ]

    # Invalid list[int]
    with pytest.raises(
        SettingMisconfigured, match="Invalid type for TEST_LIST with configured value"
    ):
        convert_and_validate("TEST_LIST", '["a", "b", "c"]', list_int_type)

    # Tuple[int, ...]
    tuple_int_type = Tuple[int, ...]
    assert convert_and_validate("TEST_TUPLE", "(1, 2, 3)", tuple_int_type) == (1, 2, 3)

    # Set[str]
    set_str_type = Set[str]
    assert convert_and_validate("TEST_SET", '{"a", "b", "c"}', set_str_type) == {
        "a",
        "b",
        "c",
    }

    # Dict[str, int]
    dict_str_int_type = Dict[str, int]
    assert convert_and_validate("TEST_DICT", '{"a": 1, "b": 2}', dict_str_int_type) == {
        "a": 1,
        "b": 2,
    }

    # Invalid Dict[str, int]
    with pytest.raises(
        SettingMisconfigured, match="Invalid type for TEST_DICT with configured value"
    ):
        convert_and_validate("TEST_DICT", '{"a": "1", "b": "2"}', dict_str_int_type)


def test_convert_and_validate_union_types():
    # Union[int, str]
    union_type = Union[int, str]
    assert convert_and_validate("TEST_UNION", "42", union_type) == 42
    assert convert_and_validate("TEST_UNION", "hello", union_type) == "hello"

    # Union with container types
    union_container_type = Union[List[int], Dict[str, int]]
    assert convert_and_validate("TEST_UNION", "[1, 2, 3]", union_container_type) == [
        1,
        2,
        3,
    ]
    assert convert_and_validate(
        "TEST_UNION", '{"a": 1, "b": 2}', union_container_type
    ) == {"a": 1, "b": 2}

    # Invalid Union
    with pytest.raises(
        SettingMisconfigured, match="Invalid type for TEST_UNION with configured value"
    ):
        convert_and_validate("TEST_UNION", "xyz", Union[int, float])


def test_convert_and_validate_custom_class():
    # Valid custom class
    assert convert_and_validate(
        "TEST_CUSTOM", "[1, 2, 3]", SimpleCustomClass
    ) == SimpleCustomClass([1, 2, 3])

    # Invalid parameter for custom class
    with pytest.raises(
        SettingMisconfigured, match="Invalid type for TEST_CUSTOM with configured value"
    ):
        convert_and_validate("TEST_CUSTOM", '["a", "b", "c"]', SimpleCustomClass)

    # Custom class with invalid method signature
    with pytest.raises(
        SettingMisconfigured,
        match="Invalid method signature for.*Expected a single parameter",
    ):
        convert_and_validate("TEST_CUSTOM", "[1, 2, 3]", InvalidCustomClass)

    # Custom class with missing type hint
    with pytest.raises(
        SettingMisconfigured, match="Invalid method signature for.*Expected a type hint"
    ):
        convert_and_validate("TEST_CUSTOM", "[1, 2, 3]", UntypedCustomClass)
